"""A simple view with a text box, to display log messages obtained from the stream handler"""

from julesTk import view

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class LogView(view.Frame):
    """A view to show log messages"""

    def __init__(self, parent):
        super(LogView, self).__init__(parent=parent)

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

        xscroll = view.tk.Scrollbar(self, orient=view.tk.HORIZONTAL)
        xscroll.grid(row=1, column=0, sticky="ew")

        yscroll = view.tk.Scrollbar(self)
        yscroll.grid(row=0, column=1, sticky="ns")

        self._text = view.tk.Text(
            self, wrap="none",
            xscrollcommand=xscroll.set,
            yscrollcommand=yscroll.set
        )
        self._text.config(state="disabled")
        self._text.bind("<1>", lambda event: self._text.focus_set())

        self._text.grid(row=0, column=0, sticky="nsew")

        xscroll.config(command=self._text.xview)
        yscroll.config(command=self._text.yview)

    @property
    def text(self):
        """The text widget"""
        return self._text

    def write(self, message):
        """Write a message to the console"""
        self._text.config(state="normal")
        self._text.insert(view.tk.END, message)
        self._text.config(state="disabled")

    def flush(self):
        pass

    def clear(self, event=None):
        """Remove all messages from the console"""
        self._text.config(state="normal")
        self._text.delete(1.0, view.tk.END)
        self._text.config(state="disabled")



"""Implementing basic view objects"""

import sys
if sys.version_info[0] < 3:
    import Tkinter as tk
    import ttk
else:
    import tkinter as tk
    from tkinter import ttk

from julesTk.app import Application

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class BaseFrame(object):
    """Empty base class used to share most basic functions between frames and views"""

    FONT_FAMILY = "Verdana"
    FONT_SMALL = (FONT_FAMILY, 8)
    FONT_NORMAL = (FONT_FAMILY, 10)
    FONT_LARGE = (FONT_FAMILY, 14)
    FONT_HUGE = (FONT_FAMILY, 16)

    @staticmethod
    def configure_grid(w, sticky="nsew", **kwargs):
        w.grid(sticky=sticky, **kwargs)

    @staticmethod
    def configure_column(w, indexes, weight=1, **kwargs):
        if not isinstance(indexes, (tuple, list)):
            indexes = [indexes]
        kwargs["weight"] = weight
        for index in indexes:
            w.grid_columnconfigure(index, **kwargs)

    @staticmethod
    def configure_row(w, indexes, weight=1, **kwargs):
        if not isinstance(indexes, (tuple, list)):
            indexes = [indexes]
        kwargs["weight"] = weight
        for index in indexes:
            w.grid_rowconfigure(index, **kwargs)


class Frame(ttk.Frame, BaseFrame):

    def __init__(self, parent):
        super(Frame, self).__init__(parent)


class BaseView(BaseFrame):
    """Base View implementing most of the view functionality
    
    Preparation and loading of widgets
    ----------------------------------
    
    Use _prepare() (abstract method) to prepare the view for showing. Including the creation of widgets. 
    Call prepare() to execute preparation.
    
    Widgets and variables can be stored in a dictionary registry.
    
    Showing, hiding and closing
    ---------------------------
    
    - call show() [ internal runs -> _show ]
    _show (abstract method) implements the specific routine, need to for showing the view.
    Call show() to execute this. show() will check if prepare() has been run before and run it necessary.
    - call hide to temporarily hide the view
    _hide() implements the specific routine for hiding the view.
    - call close to destroy the view
    _close() implements the specific routine for closing the view.
    """

    def __init__(self, parent, controller):
        """Initialize a BaseView

        :param parent: Parent view (or application)
        :param controller: Controlling controller
        """
        super(BaseView, self).__init__()
        self._configured = False
        self._parent = parent
        self._controller = controller
        self._variables = {}
        self._widgets = {}

    def __del__(self):
        self.close()

    @property
    def parent(self):
        """Return the master of this widget

        :rtype: julesTk.view.BaseView
        """
        return self._parent

    @property
    def application(self):
        """Return the root level widget

        :rtype: Tkinter.Tk or tkinter.Tk
        """
        result = self.parent
        if not isinstance(result, Application):
            result = self.parent.application
        return result

    @property
    def controller(self):
        """ The controller

        :rtype: julesTk.controller.BaseController
        """
        return self._controller

    @property
    def variables(self):
        """Variables registered to this view

        :rtype: dict[str, Tkinter.Variable | tkinter.Variable]
        """
        return self._variables

    @property
    def widgets(self):
        """Widgets registered to this view

        :rtype: dict[str, Tkinter.BaseWidget | tkinter.BaseWidget]
        """
        return self._widgets

    def is_configured(self):
        return self._configured is True

    def prepare(self):
        self._prepare()
        self._configured = True
        return self

    def _prepare(self):
        raise NotImplementedError

    def show(self):
        if not self.is_configured():
            self.prepare()
        return self._show()

    def _show(self):
        raise NotImplementedError

    def hide(self):
        return self._hide()

    def _hide(self):
        raise NotImplementedError

    def close(self):
        return self._close()

    def _close(self):
        raise NotImplementedError

    def has_variable(self, name):
        """Whether a variable is registered under the given name

        :param name: Name to search for
        :type name: str
        :rtype: bool
        """
        return name in self.variables.keys()

    def add_variable(self, name, variable):
        """Register a variable under a new name to the view

        :param name: Name to register the variable under
        :type name: str
        :param variable: The variable to register
        :type variable: Tkinter.Variable | tkinter.Variable
        :rtype: Tkinter.Variable | tkinter.Variable
        """
        if self.has_variable(name):
            raise KeyError("Variable name already registered to this view")
        self.variables[name] = variable
        return variable

    def get_variable(self, name):
        """Get the variable registered to this view under the given name

        :param name: Name of the variable
        :type name: str
        :return: The requested variable
        :rtype: Tkinter.Variable | tkinter.Variable
        """
        if not self.has_variable(name):
            raise KeyError("Variable name not registered to this view")
        return self.variables[name]

    def remove_variable(self, name):
        """Remove a variable from the registry of this view

        The variable is NOT destroyed!

        :param name: Name of the variable to remove
        :type name: str
        :return: The removed variable
        :rtype: Tkinter.Variable | tkinter.Variable
        """
        if not self.has_variable(name):
            raise KeyError("Variable name not registered to this view")
        self.variables.pop(name)
        return not self.has_variable(name)

    def has_widget(self, name):
        """Whether a widget is registered using the given name

        :type name: str
        :rtype: bool
        """
        return name in self.widgets.keys()

    def add_widget(self, name, widget):
        """Register a new widget to this view

        :param name: Name of the widget
        :type name: str
        :param widget: The widget to register
        :type widget: Tkinter.BaseWidget | tkinter.BaseWidget
        :return: The registered widget
        :rtype: Tkinter.BaseWidget | tkinter.BaseWidget
        """
        if self.has_widget(name):
            raise KeyError("Widget name already registered to this view")
        self.widgets[name] = widget
        return widget

    def get_widget(self, name):
        """Get the widget registered under the given name

        :param name: Name of the widget
        :type name: str
        :return: The requested widget
        :rtype: Tkinter.BaseWidget | tkinter.BaseWidget
        """
        if not self.has_widget(name):
            raise KeyError("Widget name not registered to this view")
        return self.widgets[name]

    def remove_widget(self, name):
        """Remove a widget from the registry of this view

        The widget is NOT destroyed

        :param name: Name of the widget to remove
        :type name: str
        :return: The widget that was removed
        :rtype: Tkinter.BaseWidget | tkinter.BaseWidget
        """
        if not self.has_widget(name):
            raise KeyError("Widget name not registered to this view")
        self.widgets.pop(name)
        return not self.has_widget(name)


class View(Frame, BaseView):
    """A Frame acting as """

    def __init__(self, parent, controller):
        Frame.__init__(self, parent)
        BaseView.__init__(self, parent, controller)

    def _show(self):
        """Shows the view"""
        self.tkraise()
        return True

    def _hide(self):
        """Hides the view (temporarily)"""
        self.grid_remove()
        return True

    def _close(self):
        """Closes the view"""
        self.destroy()
        return True

    def _prepare(self):
        """Configure this view"""
        raise NotImplementedError




from . import ViewController


class ViewSetController(ViewController):

    def __init__(self, parent, view=None):
        super(ViewSetController, self).__init__(parent, view)
        self._controllers = {}

    @property
    def controllers(self):
        """ Dictionary with all controllers used in this viewset

        :return:
        :rtype: dict[str, julesTk.controller.BaseController]
        """
        return self._controllers

    def has_controller(self, name):
        """Whether a controller is registered to this controller using the given name"""
        return name in self.controllers.keys()

    def get_controller(self, name):
        """Return the controller registered under the given name"""
        if not self.has_controller(name):
            raise KeyError("No controller registered using the name: {}".format(name))
        return self.controllers[name]

    def add_controller(self, name, c):
        """Register a controller under a new name"""
        if self.has_controller(name):
            raise KeyError("Another controller is already registered under: {}".format(name))
        self.controllers[name] = c

    def remove_controller(self, name):
        """Remove controller and name from the registry"""
        if not self.has_controller(name):
            raise KeyError("No controller registered using the name: {}".format(name))
        return self.controllers.pop(name)

    def remove_controllers(self):
        """Remove all controllers from the registry

        And tell them to stop
        """
        while len(self.controllers.keys()) > 0:
            key = self.controllers.keys()[0]
            self.remove_controller(key).stop()
        return len(self.controllers.keys()) == 0

    def _stop(self):
        self.remove_controllers()
        super(ViewSetController, self)._stop()



"""Class implementing a ProgressBar Dialog"""

from modals import view, SimpleDialog

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class ProgressBar(SimpleDialog):

    def __init__(self, parent, controller, mode="indeterminate", blocking=True):
        self._value = view.tk.DoubleVar(0)
        self._mode = mode
        self._blocking = blocking
        buttons = [{"id": "ok", "caption": "Close", "value": False}]
        super(ProgressBar, self).__init__(parent, controller, buttons=buttons)

    @property
    def bar(self):
        result = None
        if self.has_widget("progress"):
            result = self.get_widget("progress")
        return result

    def has_bar(self):
        return self.bar is not None

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, v):
        if v not in ["determinate", "indeterminate"]:
            raise ValueError("Invalid value for mode: {}".format(v))
        self._mode = v
        # update widget
        if self.has_widget("progress"):
            self.get_widget("progress").config(mode = self._mode)

    @property
    def value(self):
        return self._value.get()

    @value.setter
    def value(self, v):
        if not isinstance(v, (int, float)):
            raise ValueError("Invalid value: {}".format(v))
        self._value.set(v)
        if self.has_bar():
            maximum = self.bar.cget("maximum")
            if v >= maximum:
                self.set_blocking(False)

    def is_blocking(self):
        return self._blocking is True

    def set_blocking(self, state):
        self._blocking = state is True
        if self.has_widget("ok"):
            self.get_widget("ok").config(state="disabled" if state else "normal")

    def body(self, parent):
        super(ProgressBar, self).body(parent)
        pgb = view.ttk.Progressbar(parent, variable=self._value)
        pgb.pack(side=view.tk.BOTTOM, fill=view.tk.X, expand=1)
        self.add_widget("progress", pgb)

    def _show(self):
        if self.mode == "indeterminate":
            self.get_widget("progress").start()
        if self.is_blocking():
            self.get_widget("ok").config(state="disabled")
        super(ProgressBar, self)._show()

    def process_click(self, value):
        self._response = value
        if not self.is_blocking():
            self.close()
